import { Buff }          from '@cmdcode/buff'
import { decode_script } from '@scrow/tapscript/script'
import { hash_tx }       from '@scrow/tapscript/sighash'
import { create_tx }     from '@scrow/tapscript/tx'
import { parse_psbt }    from './parse.js'
import { assert }        from '@/util/index.js'

import { decode_prevout, decode_witness } from './tx.js'

import {
  SigHashOptions,
  TxData,
  TxInTemplate,
  TxOutTemplate,
  TxTemplate
} from '@scrow/tapscript'

import {
  collect_input_record,
  find_input_record,
  get_input_record
} from './records.js'

import {
  PSBTData,
  PSBTGlobalRecord,
  PSBTInputRecord,
  PSBTInputSegwitData,
  PSBTInputTaprootData,
  PSBTOutputRecord
} from '@/types/psbt.js'

export function extract_txdata (psbt : string | PSBTData) : TxData {
  // Unpack PSBT object.
  const { global, inputs, outputs } = parse_psbt(psbt)

  const template : TxTemplate = {}

  extract_tx_headers(global, template)

  for (let i = 0; i < inputs.length; i++) {
    assert.array_idx_exists(i, template.vin)
    template.vin[i] = extract_txin_data(inputs[i])
  }

  for (let i = 0; i < outputs.length; i++) {
    assert.array_idx_exists(i, template.vout)
    template.vout[i] = extract_txout_data(outputs[i])
  }

  return create_tx(template)
}

export function extract_input_segwit_data (
  records : PSBTInputRecord[]
) : PSBTInputSegwitData {
  const paths  = collect_input_record(records, 'BIP32_DERIVATION')
  const sigs   = collect_input_record(records, 'PARTIAL_SIG')
  const script = find_input_record(records, 'WITNESS_SCRIPT')
  return {
    hd_paths : paths.map(e => [ e.key, e.value ]),
    sigs     : sigs.map(e =>  [ e.key, e.value ]),
    script   : script?.value
  }
}

export function extract_input_taproot_data (
  records : PSBTInputRecord[]
) : PSBTInputTaprootData {
  const paths   = collect_input_record(records, 'TAP_BIP32_DERIVATION')
  const k_sigs  = collect_input_record(records, 'TAP_KEY_SIG')
  const s_sigs  = collect_input_record(records, 'TAP_SCRIPT_SIG')
  const scripts = collect_input_record(records, 'TAP_LEAF_SCRIPT')
  return {
    hd_paths    : paths.map(e =>   [ e.key, e.value ]),
    scripts     : scripts.map(e => [ e.key, e.value ]),
    key_sigs    : k_sigs.map(e =>  [ e.key, e.value ]),
    script_sigs : s_sigs.map(e =>  [ e.key, e.value ])
  }
}

export function extract_input_sighash (
  psbt     : string | PSBTData,
  index    : number,
  options ?: SigHashOptions
) {
  const opt     = options ?? {}
  const pdata   = parse_psbt(psbt)
  const input   = pdata.inputs.at(index)
  const txdata  = extract_txdata(pdata)
  const txinput = txdata.vin.at(index)
  assert.exists(input)
  assert.exists(txinput)
  opt.txindex = index
  opt.txinput = txinput
  opt.sigflag = extract_input_sigflag(input)
  return hash_tx(txdata, opt)
}

function extract_tx_headers (
  global   : PSBTGlobalRecord[],
  template : TxTemplate
) {
  for (const record of global) {
    switch (record.label) {
      case 'FALLBACK_LOCKTIME':
        template.locktime = Buff.hex(record.value).num
        break
      case 'TX_VERSION':
        template.version = Buff.hex(record.value).num
        break
      case 'INPUT_COUNT':
        template.vin = new Array(Buff.hex(record.value).num)
        break
      case 'OUTPUT_COUNT':
        template.vout = new Array(Buff.hex(record.value).num)
        break
    }
  }
}

function extract_txin_data (input : PSBTInputRecord[]) {
  const template : Partial<TxInTemplate> = {}
  for (const record of input) {
    switch (record.label) {
      case 'PREVIOUS_TXID':
        template.txid = record.value
        break
      case 'OUTPUT_INDEX':
        template.vout = Buff.hex(record.value).num
        break
      case 'SEQUENCE':
        template.sequence = record.value
        break
      case 'WITNESS_UTXO':
        template.prevout = decode_prevout(record.value)
        break
      case 'FINAL_SCRIPTWITNESS':
        template.witness = decode_witness(record.value)
        break
    }
  }
  assert.exists(template.txid)
  assert.exists(template.vout)
  return template as TxInTemplate
}

function extract_txout_data (output : PSBTOutputRecord[]) {
  const template : Partial<TxOutTemplate> = {}
  for (const record of output) {
    switch (record.label) {
      case 'AMOUNT':
        template.value = Buff.hex(record.value).big
        break
      case 'SCRIPT':
        template.scriptPubKey = decode_script(record.value, true)
        break
    }
  }
  assert.exists(template.value)
  assert.exists(template.scriptPubKey)
  return template as TxOutTemplate
}

export function extract_input_sigflag (
  records : PSBTInputRecord[]
) {
  const record = get_input_record(records, 'SIGHASH_TYPE')
  return Buff.hex(record.value).num
}

export function extract_input_prevout (
  records : PSBTInputRecord[]
) {
  const record = get_input_record(records, 'WITNESS_UTXO')
  return decode_prevout(record.value)
}
