import { parse_psbt } from './parse.js'
import { assert }     from '@/util/index.js'

import { TxData, TxInput, TxOutput, TxPrevout } from '@scrow/tapscript'

import {
  update_global_record,
  update_input_record,
  update_output_record
} from './records.js'

import {
  populate_psbt,
  populate_txin_data,
  populate_txout_data
} from './populate.js'

import {
  PSBTData,
  PSBTGlobalRecord,
  PSBTInputRecord,
  PSBTOutputRecord
} from '@/types/index.js'

import * as verify from './verify.js'

export function update_psbt_global (
  psbt     : string | PSBTData,
  template : PSBTGlobalRecord[]
) : PSBTData {
  // Unpack PSBT data object.
  const { global, inputs, outputs } = parse_psbt(psbt)
  // Iterate through each record and update the global fields.
  for (const key of template ?? []) {
    update_global_record(global, key)
  }
  // Validate and return the new PSBT.
  return { global, inputs, outputs }
}

export function update_psbt_input (
  psbt     : string | PSBTData,
  index    : number,
  template : PSBTInputRecord[]
) : PSBTData {
  // Unpack PSBT data object.
  const { global, inputs, outputs } = parse_psbt(psbt)
  // Iterate through each record and update the input fields.
  for (const key of template ?? []) {
    update_input_record(inputs[index], key)
  }
  // Validate and return the new PSBT.
  return { global, inputs, outputs }
}

export function update_psbt_output (
  psbt     : string | PSBTData,
  index    : number,
  template : PSBTOutputRecord[]
) : PSBTData {
  // Unpack PSBT data object.
  const { global, inputs, outputs } = parse_psbt(psbt)
  // Iterate through each record and update the global fields.
  for (const key of template ?? []) {
    update_output_record(outputs[index], key)
  }
  // Validate and return the new PSBT.
  return { global, inputs, outputs }
}

export function update_txdata (
  psbt   : PSBTData,
  txdata : TxData
) {
  const { global, inputs, outputs } = psbt
  // Verify the PSBT is version 2.
  verify.psbt_version(global, 2)
  // Assert the input and output lengths match.
  assert.ok(txdata.vin.length  === inputs.length,  'PSBT input length does not match unsigned tx')
  assert.ok(txdata.vout.length === outputs.length, 'PSBT output length does not match unsigned tx')
  // Update the PSBT with transaction data.
  populate_psbt({ global, inputs, outputs }, txdata)
  // Return the updated PSBT.
  return { global, inputs, outputs }
}

export function update_txin (
  psbt  : string  | PSBTData,
  txin  : TxInput | TxPrevout,
  index : number
) : PSBTData {
  // Unpack PSBT data object.
  const { global, inputs, outputs } = parse_psbt(psbt)
  // Verify the PSBT is version 2.
  verify.psbt_version(global, 2)
  // Assert that the utxo index does not exceed the PSBT input index.
  assert.ok(inputs.length > index, 'index exceeds psbt input count')
  // Add utxo records to PSBT input.
  populate_txin_data(inputs[index], txin)
  // Validate and return the new PSBT.
  return { global, inputs, outputs }
}

export function update_txout (
  psbt  : string | PSBTData,
  txout : TxOutput,
  index : number
) : PSBTData {
  // Unpack PSBT data object.
  const { global, inputs, outputs } = parse_psbt(psbt)
  // Verify the PSBT is version 2.
  verify.psbt_version(global, 2)
  // Assert that the utxo index does not exceed the PSBT input index.
  assert.ok(outputs.length > index, 'index exceeds psbt input count')
  // Add utxo records to PSBT input.
  populate_txout_data(outputs[index], txout)
  // Validate and return the new PSBT.
  return { global, inputs, outputs }
}
